import torch


def pad_x(X: torch.Tensor, num_features=100):
    if num_features is None:
        return X
    n_features = X.shape[-1]
    zero_feature_padding = torch.zeros((*X.shape[:-1], num_features - n_features), device=X.device)
    return torch.cat([X, zero_feature_padding], dim=-1)



import math
import os
from abc import ABC
from pathlib import Path
from typing import Dict, Literal

import faiss
import numpy as np
import torch
from huggingface_hub import hf_hub_download
from sklearn.decomposition import TruncatedSVD
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.preprocessing import FunctionTransformer, OneHotEncoder
from tqdm import tqdm

from .models import InContextModel
from .utils import sample_confidence_interval


def is_hf_model_path(model_path: str) -> bool:
    """
    Check if a given path is a Hugging Face model path (repo_id/model_name).

    Args:
        model_path: The model path to check

    Returns:
        bool: True if it's a Hugging Face model path, False otherwise
    """
    # Check if path doesn't exist locally but follows the pattern of org/repo or user/repo
    if not os.path.exists(model_path) and "/" in model_path and model_path.count("/") == 1:
        return True
    return False


def download_from_hf_hub(model_path: str, cache_dir: str) -> str:
    """
    Download a model from the Hugging Face Hub.

    Args:
        model_path: The model path in format 'org/repo' or 'user/repo'
        cache_dir: Optional directory to cache the downloaded model

    Returns:
        str: Path to the downloaded model file
    """
    # Download the model
    local_path = hf_hub_download(
        repo_id=model_path,
        filename="causalpfn_v0.pt",
        cache_dir=cache_dir,
    )
    return local_path


class CausalEstimator(ABC):
    """
    Shared functionalities between CATEEstimator and ATEEstimator.

    Args:
        device (str): The device to run the model on (e.g., 'cuda' or 'cpu').
        model_path (str): The path to the model checkpoint. Can be:
            - A local file path
            - A Hugging Face model path
        cache_dir (str, optional): Directory to cache downloaded models from Hugging Face.
            Defaults to ~/.cache/causalpfn.
        calibrate (bool): Whether to calibrate the model's temperature using n-fold cross-validation.
        n_folds (int): The number of folds used for cross-validation to find the best calibration.
        ICE_n_bins (int): The number of bins used for the integrated coveragre error (ICE) calculation.
        ICE_n_samples (int): The number of samples used for integrated coveragre error (ICE) calculation.
        calibrate_T_min (float): The minimum temperature to use in calibration.
        calibrate_T_max (float): The maximum temperature to use in calibration.
        calibrate_T_size (int): The number of different temperature values to search during calibration.
        calibrate_T_batch_size (int): The batch size to use for calibration (should be <= `calibrate_T_size`).
        verbose (bool): Whether to print progress messages for the `predict_cep` function.

        *NOTE*: all of the following parameters only make sense when the data is too large
                to fit inside the context length of the model.

        max_context_length (int):
            The maximum number of context examples to use for each query.
            Set to as large as your GPU can handle.
        max_query_length (int):
            The maximum number of query examples to process at once.
            Set to as large as your GPU can handle.
        num_neighbours (int):
            The number of neerest neighbours to use for the CEPO prediction
    """

    def __init__(
        self,
        device: str,
        model_path: str = "vdblm/causalpfn",
        max_context_length: int = 4096,
        max_query_length: int = 4096,
        num_neighbours: int = 1024,
        calibrate: bool = False,
        n_folds: int = 3,
        calibrate_T_min: float = 0.001,
        calibrate_T_max: float = 10,
        calibrate_T_size: int = 500,
        calibrate_T_batch_size: int = 50,
        ICE_n_bins: int = 10,
        ICE_n_samples: int = 1000,
        verbose: bool = False,
        cache_dir: str | None = None,
    ):
        self.model_path = model_path
        self.cache_dir = cache_dir if cache_dir is not None else os.path.join(Path.home(), ".cache", "causalpfn")
        self.icl_model: InContextModel = None

        self.device = device
        self.max_context_length = max_context_length
        self.max_query_length = max_query_length
        self.num_neighbours = num_neighbours

        # The maximum number of features to use for the model. If the number of features are
        # larger than this value, the model will apply PCA to reduce the dimensionality.
        self.max_feature_size = None
        self.x_dim_transformer = FunctionTransformer()  # identity transformer by default

        self.n_folds = n_folds

        self.calibrate = calibrate
        self.calibrate_T_min = calibrate_T_min
        self.calibrate_T_max = calibrate_T_max
        self.calibrate_T_size = calibrate_T_size
        self.calibrate_T_batch_size = calibrate_T_batch_size

        self.ICE_n_bins = ICE_n_bins
        self.ICE_n_samples = ICE_n_samples

        self.X_train, self.t_train, self.y_train = None, None, None
        self.temperature = 1.0
        self.prediction_temperature = 1.0

        self.verbose = verbose

    def _check_fitted(self):
        if self.X_train is None or self.t_train is None or self.y_train is None or self.icl_model is None:
            raise ValueError("The estimator must be fitted before calling the estimate function.")

    def load_model(self):
        """
        Load the model from the specified path or download it from Hugging Face.
        """
        model_path = self.model_path

        # Check if the model path is a Hugging Face model path
        if is_hf_model_path(model_path):
            model_path = download_from_hf_hub(model_path, self.cache_dir)

        # Load the model from the local path
        ckpt = torch.load(model_path, weights_only=False, map_location="cpu")
        model_state = ckpt["model_state_dict"]
        config = ckpt["model_config"]

        self.icl_model = InContextModel.load(model_state=model_state, model_config=config).to(self.device)

        if config["model_type"] == "tabdpt":
            self.max_feature_size = config["model"]["max_num_features"] - 1
            self.x_dim_transformer = TruncatedSVD(n_components=self.max_feature_size, algorithm="arpack")

    @torch.no_grad()
    def _predict_cepo(
        self,
        X_context: np.ndarray,
        t_context: np.ndarray,
        y_context: np.ndarray,
        X_query: np.ndarray,
        t_query: np.ndarray,
        temperature: float,
        n_samples: int | None = None,
    ) -> np.ndarray:
        if self.icl_model is None:
            raise ValueError("CausalEstimator must be fitted before calling _predict_cepo.")

        temperature = torch.tensor([temperature], device=self.device)
        self.icl_model: InContextModel
        self.icl_model.eval()

        # list all of the point estimates as well as distributional estimates
        # of the CEPO in all_cepo and all_samples, respectively
        all_cepo = np.zeros((X_query.shape[0],), dtype=X_query.dtype)
        if n_samples is not None:
            all_samples = np.zeros((X_query.shape[0], n_samples), dtype=X_query.dtype)

        context_effects = self._estimate_cate_weak_learner(X_test=X_context)
        treatmentgroup_2_context_idx = np.where(np.isclose(t_context, 1))[0]
        controlgroup_2_context_idx = np.where(np.isclose(t_context, 0))[0]
        context_treatment_group_effects = context_effects[treatmentgroup_2_context_idx]
        context_control_group_effects = context_effects[controlgroup_2_context_idx]

        query_effects = self._estimate_cate_weak_learner(X_test=X_query)
        query_indices_sorted = np.argsort(query_effects)

        index_treatment = faiss.IndexFlatL2(1)
        index_treatment.add(
            np.ascontiguousarray(context_treatment_group_effects.reshape(-1, 1).copy(), dtype=np.float32)
        )
        _, query_neighbour_indices_treatment = index_treatment.search(
            np.ascontiguousarray(query_effects.reshape(-1, 1).copy(), dtype=np.float32), k=self.num_neighbours
        )
        index_control = faiss.IndexFlatL2(1)
        index_control.add(np.ascontiguousarray(context_control_group_effects.reshape(-1, 1).copy(), dtype=np.float32))
        _, query_neighbour_indices_control = index_control.search(
            np.ascontiguousarray(query_effects.reshape(-1, 1).copy(), dtype=np.float32), k=self.num_neighbours
        )

        # if the query size is large, we split it into batches
        pbar = tqdm(range(X_query.shape[0]), desc="Predicting CEPO", total=X_query.shape[0], disable=not self.verbose)
        start_idx = 0
        while start_idx < X_query.shape[0]:

            def get_stratum(query_neighbour_indices):
                # binary search to find the maximum window size for the query
                # that has neighbours less than the maximum context length
                L = start_idx + 1
                R = min(start_idx + self.max_query_length, X_query.shape[0])
                while R > L + 1:
                    mid = (L + R) // 2
                    tmp_idx_q = query_indices_sorted[start_idx:mid]
                    unique_indices = np.unique(query_neighbour_indices[tmp_idx_q].flatten())
                    if len(unique_indices) > self.max_context_length // 2:
                        R = mid - 1
                    else:
                        L = mid
                return min(L, X_query.shape[0])

            control_end_idx = get_stratum(query_neighbour_indices_control)
            control_q_idx = query_indices_sorted[start_idx:control_end_idx]
            all_neighbours = query_neighbour_indices_control[control_q_idx].flatten()
            all_neighbours = np.unique(all_neighbours)
            real_control_idx = controlgroup_2_context_idx[all_neighbours]
            treatment_end_idx = get_stratum(query_neighbour_indices_treatment)
            treatment_q_idx = query_indices_sorted[start_idx:treatment_end_idx]
            all_neighbours = query_neighbour_indices_treatment[treatment_q_idx].flatten()
            all_neighbours = np.unique(all_neighbours)
            real_treatment_idx = treatmentgroup_2_context_idx[all_neighbours]

            # now use that window size to get the query and context
            end_idx = min(control_end_idx, treatment_end_idx)
            idx_q = query_indices_sorted[start_idx:end_idx]
            x_q = X_query[idx_q]
            t_q = t_query[idx_q]

            # find the closest neighbours in terms of the stratifying CATE
            all_neighbours = np.concatenate([real_control_idx, real_treatment_idx])
            assert (
                len(all_neighbours) <= self.max_context_length
            ), f"Number of neighbours {len(all_neighbours)} is larger than the maximum context length {self.max_context_length}."
            x_c = X_context[all_neighbours]
            t_c = t_context[all_neighbours]
            y_c = y_context[all_neighbours]

            res = self.icl_model.predict_cepo(
                # shape: (1, context_size, num_features)
                X_context=torch.from_numpy(x_c).to(self.device).unsqueeze(0).float(),
                # shape: (1, context_size)
                t_context=torch.from_numpy(t_c).to(self.device).unsqueeze(0).float(),
                y_context=torch.from_numpy(y_c).to(self.device).unsqueeze(0).float(),
                # shape: (1, query_size, num_features)
                X_query=torch.from_numpy(x_q).to(self.device).unsqueeze(0).float(),
                # shape: (1, query_size)
                t_query=torch.from_numpy(t_q).to(self.device).unsqueeze(0).float(),
                n_samples=n_samples,
                temperature=temperature,
            )
            if n_samples is None:
                cepo = res.squeeze(0).squeeze(0)  # shape: (query_size,)
            else:
                cepo, samples = (
                    res[0].squeeze(0).squeeze(0),
                    res[1].squeeze(0).squeeze(0),
                )  # shapes: (query_size,), (query_size, n_samples)

            all_cepo[idx_q] = cepo.cpu().numpy()
            if n_samples is not None:
                all_samples[idx_q] = samples.cpu().numpy()
            pbar.update(end_idx - start_idx)
            start_idx = end_idx

        if n_samples is not None:
            return all_cepo, all_samples

        return all_cepo

    @torch.no_grad()
    def _calculate_reg_ice(
        self,
        X: torch.Tensor,
        t: torch.Tensor,
        y: torch.Tensor,
        temperature: torch.Tensor,
    ) -> torch.Tensor:

        alpha_values = torch.linspace(0.01, 1.0, self.ICE_n_bins, device=self.device)  # significance levels
        ce_values = torch.zeros_like(temperature, device=self.device)

        fold_indices = np.array_split(np.arange(len(X)), self.n_folds)
        for val_idx in fold_indices:
            train_idx = np.setdiff1d(np.arange(len(X)), val_idx)

            X_context, t_context, y_context = X[train_idx], t[train_idx], y[train_idx]
            X_query, t_query, y_query = X[val_idx], t[val_idx], y[val_idx]

            b_size = math.ceil(X_query.shape[0] / self.max_query_length)
            batch_ice_values = torch.zeros_like(temperature, device=self.device)
            for b in range(b_size):
                batch_start = b * self.max_query_length
                batch_end = min((b + 1) * self.max_query_length, X_query.shape[0])
                if batch_start >= batch_end:
                    continue
                X_q = X_query[batch_start:batch_end]
                t_q = t_query[batch_start:batch_end]
                y_q = y_query[batch_start:batch_end]  # shape: (len(X_q), )

                if t_context.shape[0] > self.max_context_length:
                    # sample random set of context points
                    context_indices = np.random.choice(t_context.shape[0], self.max_context_length, replace=False)
                else:
                    context_indices = np.arange(t_context.shape[0])

                res = self.icl_model.predict_cepo(
                    # shape: (1, context_size, num_features)
                    X_context=X_context[context_indices].unsqueeze(0).float(),
                    # shape: (1, context_size)
                    t_context=t_context[context_indices].unsqueeze(0).float(),
                    y_context=y_context[context_indices].unsqueeze(0).float(),
                    # shape: (1, query_size, num_features)
                    X_query=X_q.unsqueeze(0).float(),
                    # shape: (1, query_size)
                    t_query=t_q.unsqueeze(0).float(),
                    n_samples=self.ICE_n_samples,
                    temperature=temperature,
                )
                samples = res[1].squeeze(0)  # shape: (len(temperature), len(X_q), n_samples)

                # shape: (len(alpha_values, len(temperature), len(X_q))
                lower_bound, upper_bound = sample_confidence_interval(samples, alphas=alpha_values)
                # shape: (len(alpha_values), len(temperature))
                coverage = (
                    ((y_q[None, None, :] >= lower_bound) & (y_q[None, None, :] <= upper_bound)).float().mean(axis=-1)
                )

                batch_ice_values += torch.abs(coverage - (1 - alpha_values[:, None])).mean(axis=0)

            ce_values += batch_ice_values / b_size

        return ce_values / self.n_folds

    def _calibrate_fn(self, X: np.ndarray, t: np.ndarray, y: np.ndarray):
        """
        Calibrate the temperature using n-fold cross-validation.
        """

        X = torch.from_numpy(X).to(self.device)
        t = torch.from_numpy(t).to(self.device)
        y = torch.from_numpy(y).to(self.device)

        temperatures = torch.logspace(
            start=np.log10(self.calibrate_T_min),
            end=np.log10(self.calibrate_T_max),
            steps=self.calibrate_T_size,
            device=self.device,
        )

        all_losses = []
        batch_size = self.calibrate_T_batch_size
        num_batches = math.ceil(len(temperatures) / batch_size)

        pbar = tqdm(range(num_batches), desc="Calibrating temperature", disable=not self.verbose)
        for i in pbar:
            start_idx = i * batch_size
            end_idx = min((i + 1) * batch_size, len(temperatures))
            batch_temperatures = temperatures[start_idx:end_idx]

            batch_losses = self._calculate_reg_ice(X, t, y, temperature=batch_temperatures)
            all_losses.append(batch_losses)
            pbar.update(i)

        losses = torch.cat(all_losses, dim=0)
        best_idx = torch.argmin(losses).item()
        self.temperature = temperatures[best_idx].item()

    def _train_weak_learner(self, X: np.ndarray, t: np.ndarray, y: np.ndarray) -> GradientBoostingRegressor:
        self.t_transformer = OneHotEncoder(sparse_output=False, categories="auto", drop="first")
        T = self.t_transformer.fit_transform(t.reshape(-1, 1))
        self._d_t = (T.shape[1],)
        feat_arr = np.concatenate((X, 1 - np.sum(T, axis=1).reshape(-1, 1), T), axis=1)
        self.stratifier = GradientBoostingRegressor(
            n_estimators=100,
            max_depth=6,
            min_samples_leaf=int(X.shape[0] / 100),
            random_state=111,
        )
        self.stratifier.fit(feat_arr, y)

    def _estimate_cate_weak_learner(self, X_test: np.ndarray) -> np.ndarray:
        if self.stratifier is None or self.t_transformer is None:
            raise ValueError("Weak learner must be trained before estimating CATE.")

        t0 = np.zeros(X_test.shape[0], dtype=X_test.dtype)
        t1 = np.ones(X_test.shape[0], dtype=X_test.dtype)
        T0 = self.t_transformer.transform(t0.reshape(-1, 1))
        T1 = self.t_transformer.transform(t1.reshape(-1, 1))
        X_T0 = np.concatenate((X_test, 1 - np.sum(T0, axis=1).reshape(-1, 1), T0), axis=1)
        X_T1 = np.concatenate((X_test, 1 - np.sum(T1, axis=1).reshape(-1, 1), T1), axis=1)
        Y0 = self.stratifier.predict(X_T0)
        Y1 = self.stratifier.predict(X_T1)

        return Y1 - Y0

    def fit(self, X: np.ndarray, t: np.ndarray, y: np.ndarray) -> "CausalEstimator":
        """
        Fit the model using the provided data.

        Args:
            X (np.ndarray): The observational covariate data with shape [N, D].
            t (np.ndarray): The observational treatment data with shape [N].
            y (np.ndarray): The observational outcome data with shape [N].
        """
        self.temperature = 1.0

        # load the model
        self.load_model()

        # set the x_dim_transform and transform the data
        if self.max_feature_size is not None and X.shape[1] > self.max_feature_size:
            X = self.x_dim_transformer.fit_transform(X)

        self.X_train = X
        self.t_train = t
        self.y_train = y

        # train a lightweight stratification model to use whenever the context length is too large
        # this ensures that the data will be stratified and the context would only look at the most relevant samples
        # for the given query
        self._train_weak_learner(X, t, y)

        if self.calibrate:
            self._calibrate_fn(
                X=self.X_train,
                t=self.t_train,
                y=self.y_train,
            )
        return self


class CATEEstimator(CausalEstimator):
    def estimate_cate(self, X: np.ndarray) -> np.ndarray:
        """
        Estimate the conditional average treatment effect (CATE) using the fitted model.
        Args:
            X (np.ndarray): The input data with shape [N', D].
        """
        self._check_fitted()

        X_context = self.X_train
        t_context = self.t_train
        y_context = self.y_train
        X_query = X
        if self.max_feature_size is not None and X_query.shape[1] > self.max_feature_size:
            X_query = self.x_dim_transformer.transform(X_query)

        t_all_ones = np.ones(X_query.shape[0], dtype=X_query.dtype)
        t_all_zeros = np.zeros(X_query.shape[0], dtype=X_query.dtype)

        mu_0_and_1 = self._predict_cepo(
            X_context=X_context,
            t_context=t_context,
            y_context=y_context,
            X_query=np.concatenate([X_query, X_query], axis=0),
            t_query=np.concatenate([t_all_zeros, t_all_ones], axis=0),
            temperature=self.prediction_temperature,
        )

        mu_0 = mu_0_and_1[: X_query.shape[0]]
        mu_1 = mu_0_and_1[X_query.shape[0] :]
        return mu_1 - mu_0

    def _estimate_ate_cate_CI(
        self, X: np.ndarray, alpha: float = 0.05, n_samples: int = 10_000
    ) -> Dict[str, np.ndarray]:
        self._check_fitted()

        X_context = self.X_train
        t_context = self.t_train
        y_context = self.y_train
        X_query = X
        if self.max_feature_size is not None and X_query.shape[1] > self.max_feature_size:
            X_query = self.x_dim_transformer.transform(X_query)

        t_all_ones = np.ones(X_query.shape[0], dtype=X_query.dtype)
        t_all_zeros = np.zeros(X_query.shape[0], dtype=X_query.dtype)

        _, samples = self._predict_cepo(
            X_context=X_context,
            t_context=t_context,
            y_context=y_context,
            X_query=np.concatenate([X_query, X_query], axis=0),
            t_query=np.concatenate([t_all_zeros, t_all_ones], axis=0),
            n_samples=n_samples,
            temperature=self.temperature,
        )

        # calculate the confidence intervals
        samples_0 = samples[: X_query.shape[0]]
        samples_1 = samples[X_query.shape[0] :]

        cate_samples = samples_1 - samples_0

        lower_bound, upper_bound = sample_confidence_interval(
            torch.from_numpy(cate_samples).float(), alphas=torch.tensor([alpha]).float()
        )

        ate_samples = samples_1.mean(axis=0) - samples_0.mean(axis=0)
        ate_lower_bound, ate_upper_bound = sample_confidence_interval(
            torch.from_numpy(ate_samples).float(), alphas=torch.tensor([alpha]).float()
        )
        return {
            "cate_lower_bound": lower_bound.numpy(),
            "cate_upper_bound": upper_bound.numpy(),
            "ate_lower_bound": ate_lower_bound.numpy(),
            "ate_upper_bound": ate_upper_bound.numpy(),
        }

    def estimate_cate_CI(self, X: np.ndarray, alpha: float = 0.05, n_samples: int = 10_000) -> Dict[str, np.ndarray]:
        """
        Estimate the conditional average treatment effect (CATE) with confidence intervals using the fitted model.

        Args:
            X (np.ndarray): The input data with shape [N', D].
            alpha (float): The significance level for the confidence interval.
            n_samples (int): The number of samples to use for estimating the confidence interval.

        Returns:
            Dict[str, np.ndarray]: A dictionary containing the confidence intervals.
                - "lower_bound": The lower bound of the confidence interval.
                - "upper_bound": The upper bound of the confidence interval.
        """
        output = self._estimate_ate_cate_CI(X, alpha=alpha, n_samples=n_samples)
        return {
            "lower_bound": output["cate_lower_bound"],
            "upper_bound": output["cate_upper_bound"],
        }

    def estimate_ate(self, X: np.ndarray) -> float:
        """
        Estimate the average treatment effect (ATE) using the fitted model and CATE values.
        This is different from the ATEEstimator in that it just simply computes the CATE values and averages them.
        """
        cate_pred = self.estimate_cate(X)
        return cate_pred.mean()

    def estimate_ate_CI(self, X: np.ndarray, alpha: float = 0.05, n_samples: int = 10_000) -> Dict[str, np.ndarray]:
        """
        Estimate the average treatment effect (ATE) with confidence intervals using the fitted model and CATE values.
        This is different from the ATEEstimator in that it just simply computes the CATE values and averages them.

        Args:
            X (np.ndarray): The input data with shape [N', D].
            alpha (float): The significance level for the confidence interval.
            n_samples (int): The number of samples to use for estimating the confidence interval.

        Returns:
            Dict[str, np.ndarray]: A dictionary containing the ATE estimates and the confidence intervals.
                - "ate": The ATE estimates.
                - "lower_bound": The lower bound of the confidence interval.
                - "upper_bound": The upper bound of the confidence interval.
        """
        output = self._estimate_ate_cate_CI(X, alpha=alpha, n_samples=n_samples)
        return {
            "ate": output["ate"],
            "lower_bound": output["ate_lower_bound"],
            "upper_bound": output["ate_upper_bound"],
        }


class ATEEstimator(CausalEstimator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mu_0 = None
        self.mu_1 = None

    def _estimate_ate_ra_learner(self) -> float:

        mu_1_final = np.where(np.isclose(self.t_train, 1), self.y_train, self.mu_1)
        mu_0_final = np.where(np.isclose(self.t_train, 0), self.y_train, self.mu_0)

        return (mu_1_final - mu_0_final).mean()

    def _estimate_ate_s_learner(self) -> float:
        return (self.mu_1 - self.mu_0).mean()

    def estimate_ate(self, algorithm: Literal["RA", "S"] = "S") -> float:
        """
        Estimate the average treatment effect (ATE) using the fitted model.
        Args:
            algorithm (str):
                When set to "S", it simply uses the model-output CEPO values,
                When set to "RA", it uses the model only to get access to the counterfactuals
        """
        self._check_fitted()
        if self.mu_0 is None or self.mu_1 is None:
            t_all_ones = np.ones(self.X_train.shape[0], dtype=self.X_train.dtype)
            t_all_zeros = np.zeros(self.X_train.shape[0], dtype=self.X_train.dtype)

            mu_0_and_1 = self._predict_cepo(
                X_context=self.X_train,
                t_context=self.t_train,
                y_context=self.y_train,
                X_query=np.concatenate([self.X_train, self.X_train], axis=0),
                t_query=np.concatenate([t_all_zeros, t_all_ones], axis=0),
                temperature=self.prediction_temperature,
            )
            self.mu_0 = mu_0_and_1[: self.X_train.shape[0]]
            self.mu_1 = mu_0_and_1[self.X_train.shape[0] :]

        if algorithm == "S":
            return self._estimate_ate_s_learner()
        elif algorithm == "RA":
            return self._estimate_ate_ra_learner()
        else:
            raise ValueError(f"Unknown algorithm: {algorithm}. Supported algorithms are 'RA', and 'S'.")



from typing import Tuple

import numpy as np

from .causal_estimator import CATEEstimator


def calculate_pehe(cate_true: np.ndarray, cate_pred: np.ndarray) -> float:
    """
    Calculate the PEHE (Precision in Estimation of Heterogeneous Effect) metric.
    """

    return np.sqrt(np.mean((cate_true - cate_pred) ** 2))


def calculate_calibration_curve_scores(
    *,
    fitted_estimator: CATEEstimator,
    X_test: np.ndarray,
    true_cate: np.ndarray,
    n_bins: int,
) -> Tuple[np.ndarray, float]:
    """
    Calculates the calibration curve and the Integrated Coverage Error (ICE).

    Returns:
        - coverage: The coverage of the calibration curve with shape (n_bins,).
        - ICE
    """

    alpha_values = np.linspace(0.01, 1.0, n_bins)  # significance levels

    X_test = X_test
    cates = true_cate  # shape: (n_samples,)

    coverage = np.zeros((len(alpha_values),))

    # Compute the calibration curve
    for i, alpha in enumerate(alpha_values):
        res = fitted_estimator.estimate_cate_CI(X=X_test, alpha=alpha, n_samples=5000)
        # 1 - alpha confidence interval
        lower_bound, upper_bound = res["lower_bound"], res["upper_bound"]  # shape: (n_samples,)
        # Compute the coverage
        coverage[i] = np.mean((cates >= lower_bound) & (cates <= upper_bound))

    # Compute ICEs
    ice = np.mean(coverage - (1 - alpha_values))

    return coverage, ice


def get_qini_curve(
    rct_treatments: np.ndarray,
    rct_outcomes: np.ndarray,
    estimated_cate: np.ndarray,
    normalize: bool = True,
) -> np.ndarray:
    """
    Calculate the Qini curve based on the estimated CATE values.
    See https://proceedings.mlr.press/v67/gutierrez17a/gutierrez17a.pdf
    """
    sorted_indices = np.argsort(estimated_cate)[::-1]

    treated_size = rct_treatments[sorted_indices].cumsum()
    control_size = (1 - rct_treatments[sorted_indices]).cumsum()
    treated_outcomes = (rct_outcomes[sorted_indices] * rct_treatments[sorted_indices]).cumsum()
    control_outcomes = (rct_outcomes[sorted_indices] * (1 - rct_treatments[sorted_indices])).cumsum()

    qini_curve = np.where(control_size != 0, treated_outcomes - control_outcomes * (treated_size / control_size), 0)
    if normalize:
        qini_curve /= np.abs(qini_curve[-1]) + 1e-8

    sum_qini = np.sum(qini_curve) / len(qini_curve)
    return qini_curve, sum_qini



from .causal_estimator import ATEEstimator, CATEEstimator
from .evaluation import calculate_calibration_curve_scores, calculate_pehe

__all__ = [
    "CATEEstimator",
    "ATEEstimator",
    "calculate_pehe",
    "calculate_calibration_curve_scores",
]



from enum import Enum

import numpy as np
import torch
import torch.nn as nn

from .transformer_layer import TransformerEncoderLayer


class Task(Enum):
    REG = 1
    CLS = 2


def maskmean(x, mask, dim):
    x = torch.where(mask, x, 0)
    return x.sum(dim=dim, keepdim=True) / mask.sum(dim=dim, keepdim=True)


def maskstd(x, mask, dim=0):
    num = mask.sum(dim=dim, keepdim=True)
    mean = maskmean(x, mask, dim=0)
    diffs = torch.where(mask, mean - x, 0)
    return ((diffs**2).sum(dim=0, keepdim=True) / (num - 1)) ** 0.5


def normalize_data(data, eval_pos):
    X = data[:eval_pos] if eval_pos > 0 else data
    mask = ~torch.isnan(X)
    mean = maskmean(X, mask, dim=0)
    std = maskstd(X, mask, dim=0) + 1e-6
    data = (data - mean) / std
    return data


def clip_outliers(data, eval_pos, n_sigma=4):
    assert len(data.shape) == 3, "X must be T,B,H"
    X = data[:eval_pos] if eval_pos > 0 else data
    mask = ~torch.isnan(X)
    mean = maskmean(X, mask, dim=0)
    cutoff = n_sigma * maskstd(X, mask, dim=0)
    mask &= cutoff >= torch.abs(X - mean)
    cutoff = n_sigma * maskstd(X, mask, dim=0)
    return torch.clip(data, mean - cutoff, mean + cutoff)


def convert_to_torch_tensor(input):
    if isinstance(input, np.ndarray):
        return torch.from_numpy(input)
    elif torch.is_tensor(input):
        return input
    else:
        raise TypeError("Input must be a NumPy array or a PyTorch tensor.")


class TabDPTLongContextModel(nn.Module):
    def __init__(
        self,
        dropout: float,
        n_out: int,
        nhead: int,
        nhid: int,
        ninp: int,
        nlayers: int,
        num_features: int,
        nbins=int,
    ):
        super().__init__()
        self.n_out = n_out
        self.ninp = ninp
        self.nbins = nbins
        self.nhead = nhead
        self.nhid = nhid
        self.nlayers = nlayers
        self.transformer_encoder = nn.ModuleList(
            [
                TransformerEncoderLayer(
                    embed_dim=ninp,
                    num_heads=nhead,
                    ff_dim=nhid,
                )
                for _ in range(nlayers)
            ]
        )
        self.num_features = num_features
        self.encoder = nn.Linear(num_features, ninp, bias=True)
        self.dropout = nn.Dropout(p=dropout)
        self.y_encoder = nn.Linear(1, ninp, bias=True)
        self.head = nn.Sequential(
            nn.Linear(ninp, nhid, bias=False), nn.GELU(), nn.Linear(nhid, n_out + nbins, bias=False)
        )
        self.xnorm = nn.LayerNorm(ninp, bias=False)
        self.ynorm = nn.LayerNorm(ninp, bias=False)

    def forward(
        self,
        x_src: torch.Tensor,
        y_src: torch.Tensor,
        return_log_act_norms: bool = False,
    ) -> torch.Tensor:
        context_length = y_src.shape[0]
        B = y_src.shape[1]
        x_src = normalize_data(x_src, -1 if self.training else context_length)
        x_src = clip_outliers(x_src, -1 if self.training else context_length, n_sigma=10)
        x_src = torch.nan_to_num(x_src, nan=0)

        x_src = self.xnorm(self.encoder(x_src))
        y_src = self.ynorm(self.y_encoder(y_src.unsqueeze(-1)))
        train_x = x_src[:context_length] + y_src
        src = torch.cat([train_x, x_src[context_length:]], 0)

        log_act_norms = {}
        log_act_norms["y"] = torch.norm(y_src, dim=-1).mean()
        log_act_norms["x"] = torch.norm(x_src, dim=-1).mean()

        for l, layer in enumerate(self.transformer_encoder):
            if l in [0, 1, 3, 6, 9]:
                log_act_norms[f"layer_{l}"] = torch.norm(src, dim=-1).mean()
            src = layer(src, context_length)

        pred = self.head(src)

        pred = 30 * torch.tanh(pred / (7.5 * src.size(-1) ** 0.5))
        if return_log_act_norms:
            return pred[context_length:], log_act_norms
        else:
            return pred[context_length:]

    @classmethod
    def load(cls, model_state, config):
        assert config.model.max_num_classes > 2
        model = TabDPTLongContextModel(
            dropout=config.training.dropout,
            n_out=config.model.max_num_classes,
            nhead=config.model.nhead,
            nhid=config.model.emsize * config.model.nhid_factor,
            ninp=config.model.emsize,
            nlayers=config.model.nlayers,
            num_features=config.model.max_num_features,
            nbins=config.model.nbins,
        )

        module_prefix = "_orig_mod."
        model_state = {k.replace(module_prefix, ""): v for k, v in model_state.items()}
        model.load_state_dict(model_state)
        model.eval()
        return model



import math
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from .loading import load_pretrained_tabdpt_model
from .model import TabDPTLongContextModel
from .utils import pad_x


class InContextModel(nn.Module):

    def __init__(
        self,
        model: TabDPTLongContextModel,
        model_config: dict,  # the config containing the constructor arguments for the model
        sigma: float = 0.5,
        vmin: float = -10.0,
        vmax: float = 10.0,
    ):
        super().__init__()
        self.model: nn.Module = model
        self.model_config = model_config

        # self.prepare_input is to be called for each of the models to do any model-specific preprocessing
        self.prepare_input = lambda x, y: (pad_x(x, model.num_features), y)
        self.nbins = model_config["model"]["nbins"]
        model_config["model_type"] = "tabdpt"
        model_config["sigma"] = sigma
        self.sigma = sigma

        # NOTE: These variables are stored to avoid re-initializing them for each forward pass
        self.vmin = vmin
        self.vmax = vmax

        bin_edges = torch.linspace(self.vmin, self.vmax, self.nbins + 1)
        bin_width = bin_edges[1] - bin_edges[0]
        bin_centers = bin_edges[:-1] + 0.5 * bin_width  # shape: (nbins,)

        self.register_buffer("bin_edges", bin_edges)  # (nbins+1,)
        self.register_buffer("bin_width", bin_width)  # () – 0-D tensor
        self.register_buffer("bin_centers", bin_centers)  # (nbins,)

    def _predict_mean(self, logits: torch.Tensor):
        probs = F.softmax(logits, dim=-1)
        return torch.sum(probs * self.bin_centers, dim=-1)

    def _sample_from_logits(self, logits: torch.Tensor, n_samples: int) -> torch.Tensor:
        # Flatten trailing dim to (..., nbins)
        orig_shape = logits.shape[:-1]

        # Convert logits to probabilities
        probs = torch.softmax(logits, dim=-1)  # shape: (..., nbins)

        # Reshape for sampling: (batch_size, nbins)
        logits_reshaped = probs.reshape(-1, self.nbins)

        # Sample indices
        sampled_indices = torch.multinomial(
            logits_reshaped, num_samples=n_samples, replacement=True
        )  # (batch_size, n_samples)

        # Convert indices to values using bin_centers
        samples = self.bin_centers[sampled_indices]  # (batch_size, n_samples)

        # Reshape back to (..., n_samples)
        samples = samples.view(*orig_shape, n_samples)

        return samples

    def _hl_gaussian_cross_entropy_loss(
        self,
        logits: torch.Tensor,
        y_target: torch.Tensor,
        sigma: float,
    ) -> torch.Tensor:
        """
        Calculate the cross-entropy loss between the predicted distribution and the target tensor.
        Args:
        logits: Tensor of shape (..., nbins) containing the predicted distribution
        y_target: Tensor of shape (...) containing target distribution
        vmin: Minimum value of the distribution
        vmax: Maximum value of the distribution
        sigma: Standard deviation for the Gaussian smoothing

        Note: This function construct a soft target distribution (one per element in y_target)
        by integrating a Gaussian centered at y_i over each bin. (Histogram Loss Gaussian or HL-Gauss)
        """
        assert sigma > 0, "Sigma must be positive."
        assert (
            y_target.shape == logits.shape[:-1]
        ), "y_target must have the same shape as logits except for the last dimension."
        # We'll compute a distribution by integrating the Gaussian cdf in each bin
        # so for bin k, the probability is cdf(upper_edge) - cdf(lower_edge).
        #   cdf(x) = 0.5 * [1 + erf((x - mu) / (sqrt(2)*sigma))]

        # Expand y_target so we can broadcast:
        # y_target: (...) => (..., 1) so we can compare with each bin center
        y_target_expanded = y_target.unsqueeze(-1)  # => (..., 1)

        # We'll use the normal CDF in a piecewise manner:
        def normal_cdf(x, mean, std):
            # cdf = 0.5 * [1 + erf((x - mean)/(sqrt(2)*std))]
            return 0.5 * (1.0 + torch.erf((x - mean) / (math.sqrt(2) * std)))

        # lower and upper edges for each bin, shape (nbins,)
        lower_edges = self.bin_centers - 0.5 * self.bin_width
        upper_edges = self.bin_centers + 0.5 * self.bin_width

        # Now we want to do cdf(upper_edges) - cdf(lower_edges) for each data point
        # We'll broadcast them so that shape => (..., nbins)
        cdf_upper = normal_cdf(upper_edges, y_target_expanded, sigma)
        cdf_lower = normal_cdf(lower_edges, y_target_expanded, sigma)

        # Probability in each bin is difference of cdfs
        p = cdf_upper - cdf_lower  # shape: (..., nbins)

        # Because of numerical issues, it might not sum exactly to 1, so we can renormalize:
        p = p / (p.sum(dim=-1, keepdim=True) + 1e-8)  # shape: (..., nbins)

        # Now we can compute the cross-entropy loss with the *soft* target:  CE = - sum_k [ p_k * log(q_k) ]
        log_probs = F.log_softmax(logits, dim=-1)  # shape: (..., nbins)
        ce_loss = -torch.sum(p * log_probs, dim=-1).mean(dim=-1)
        return ce_loss

    def _get_y0_y1_shift_scale(
        self,
        t_context: torch.Tensor,
        y_context: torch.Tensor,
    ):
        # compute the mean and std of the treatment and control groups in the observational data in the context
        treated_mask = (t_context == 1).long()
        control_mask = (t_context == 0).long()
        treated_counts = treated_mask.sum(dim=1, keepdim=True)
        control_counts = control_mask.sum(dim=1, keepdim=True)

        y1_mean = (y_context * treated_mask).sum(dim=1, keepdim=True) / treated_counts
        y0_mean = (y_context * control_mask).sum(dim=1, keepdim=True) / control_counts
        y1_std = (
            torch.sqrt(((y_context - y1_mean) ** 2 * treated_mask).sum(dim=1, keepdim=True) / treated_counts) + 1e-20
        )
        y0_std = (
            torch.sqrt(((y_context - y0_mean) ** 2 * control_mask).sum(dim=1, keepdim=True) / control_counts) + 1e-20
        )

        # compute the shift and scale for the treatment and control groups
        # if no treatment or control samples are present, we set the shift and scale according to the observational data
        y0_scale = torch.where(control_counts > 0, y0_std, y1_std)
        y1_scale = torch.where(treated_counts > 0, y1_std, y0_std)
        y0_shift = torch.where(control_counts > 0, y0_mean, y1_mean)
        y1_shift = torch.where(treated_counts > 0, y1_mean, y0_mean)

        return y0_shift, y0_scale, y1_shift, y1_scale

    def cepo_losses(
        self,
        X_context: torch.Tensor,
        t_context: torch.Tensor,
        y_context: torch.Tensor,
        X_query: torch.Tensor,
        E_y0_query: torch.Tensor,
        E_y1_query: torch.Tensor,
        sigma: float | None = None,
        temperature: float = 1.0,
    ) -> torch.Tensor:
        sigma = sigma or self.sigma
        y0_shift, y0_scale, y1_shift, y1_scale = self._get_y0_y1_shift_scale(t_context, y_context)

        # z-standardize the outcomes
        y_standardized = torch.where(
            t_context == 1, (y_context - y1_shift) / y1_scale, (y_context - y0_shift) / y0_scale
        )
        E_y1_standardized = (E_y1_query - y1_shift) / y1_scale
        E_y0_standardized = (E_y0_query - y0_shift) / y0_scale

        # get access to both the factual and counterfactual predictions by evaluating each of the query points at
        # both t=1 and t=0
        random_treatments = torch.randint(0, 2, E_y0_query.shape, device=E_y0_query.device)

        x_and_t_context = torch.cat(
            [
                t_context.unsqueeze(-1),
                X_context,
            ],
            dim=2,
        )  # shape: (batch_size,  context_len , num_features + 1)

        x_and_t_query = torch.cat(
            [
                random_treatments.unsqueeze(-1),
                X_query,
            ],
            dim=2,
        )  # shape: (batch_size,  query_len , num_features + 1)

        Ey_target = torch.where(
            random_treatments == 0, E_y0_standardized, E_y1_standardized
        )  # shape: (batch_size, query_len)

        x_and_t = torch.cat(
            [x_and_t_context, x_and_t_query], dim=1
        )  # shape: (batch_size, context_len + query_len, num_features + 1)
        x_src, y_src = self.prepare_input(x_and_t, y_standardized)

        logits = self.model(x_src.transpose(0, 1), y_src.transpose(0, 1)).transpose(
            0, 1
        )  # shape: (batch_size, query_len, nbins)
        logits = logits[:, :, -self.model.nbins :]  # only keep the last nbins, which are the predictions

        logits /= temperature  # Apply temperature scaling
        return self._hl_gaussian_cross_entropy_loss(
            logits=logits,
            y_target=Ey_target,
            sigma=sigma,
        )

    def forward(
        self,
        X_context: torch.Tensor,
        t_context: torch.Tensor,
        y_context: torch.Tensor,
        X_query: torch.Tensor,
        E_y0_query: torch.Tensor,
        E_y1_query: torch.Tensor,
        sigma: float | None = None,
        temperature: float = 1.0,
    ):
        """
        The forward method will simply call the cepo_losses method and returns the loss for training.
        This is done to support multi-GPU training.
        """
        return self.cepo_losses(
            X_context=X_context,
            t_context=t_context,
            y_context=y_context,
            X_query=X_query,
            E_y0_query=E_y0_query,
            E_y1_query=E_y1_query,
            sigma=sigma,
            temperature=temperature,
        )

    def predict_cepo(
        self,
        X_context: torch.Tensor,
        t_context: torch.Tensor,
        y_context: torch.Tensor,
        X_query: torch.Tensor,
        t_query: torch.Tensor,
        n_samples: int | None = None,
        temperature: torch.Tensor | None = None,  # shape: (num_temperatures, )
    ) -> torch.Tensor | Tuple[torch.Tensor, torch.Tensor]:

        y0_shift, y0_scale, y1_shift, y1_scale = self._get_y0_y1_shift_scale(t_context, y_context)
        y_standardized = torch.where(
            t_context == 1, (y_context - y1_shift) / y1_scale, (y_context - y0_shift) / y0_scale
        )

        x_and_t_context = torch.cat(
            [
                t_context.unsqueeze(-1),
                X_context,
            ],
            dim=2,
        )  # shape: (batch_size,  context_len , num_features + 1)

        x_and_t_query = torch.cat(
            [
                t_query.unsqueeze(-1),
                X_query,
            ],
            dim=2,
        )  # shape: (batch_size,  query_len , num_features + 1)

        x_and_t = torch.cat(
            [x_and_t_context, x_and_t_query], dim=1
        )  # shape: (batch_size, context_len + query_len, num_features + 1)

        x_src, y_src = self.prepare_input(x_and_t, y_standardized)

        logits = self.model(x_src.transpose(0, 1), y_src.transpose(0, 1)).transpose(
            0, 1
        )  # shape: (batch_size, query_len, nbins)
        logits = logits[:, :, -self.model.nbins :]  # only keep the last nbins, which are the predictions

        logits = logits.unsqueeze(1)  # shape: (batch_size, 1, query_len, nbins)
        y1_scale, y1_shift = y1_scale.unsqueeze(1), y1_shift.unsqueeze(1)  # shape: (batch_size, 1, query_len)
        y0_scale, y0_shift = y0_scale.unsqueeze(1), y0_shift.unsqueeze(1)  # shape: (batch_size, 1, query_len)
        t_query = t_query.unsqueeze(1)  # shape: (batch_size, 1, query_len)

        if temperature is None:
            logits = logits.unsqueeze(1)
        if temperature is not None:
            temperature = temperature[None, :, None, None]  # shape: (1, num_temperatures, 1, 1)

        logits = logits / temperature  # Apply temperature scaling

        mean = self._predict_mean(logits)  # shape: (batch_size, num_temperatures, query_len)
        mean_shift_scaled = torch.where(t_query == 1, mean * y1_scale + y1_shift, mean * y0_scale + y0_shift)

        if n_samples is None:
            return mean_shift_scaled

        # shape: (batch_size, num_temperatures, query_len, n_samples)
        samples = self._sample_from_logits(logits, n_samples)
        samples_shift_scaled = torch.where(
            t_query.unsqueeze(-1) == 1, samples * y1_scale + y1_shift, samples * y0_scale + y0_shift
        )
        return mean_shift_scaled, samples_shift_scaled

    @classmethod
    def load(cls, model_state: dict, model_config: dict) -> "InContextModel":
        inner_model_state = {k.replace("model.", ""): v for k, v in model_state.items() if k.startswith("model.")}
        if model_config["model_type"] == "tabdpt":
            ckpt_loaded = {"cfg": model_config, "model": inner_model_state}
            base_model = load_pretrained_tabdpt_model(ckpt_path=None, ckpt=ckpt_loaded)
        else:
            raise ValueError(f"Unknown model. Supported model is 'tabdpt'.")

        sigma = model_config.get("sigma", 0.5)
        return InContextModel(model=base_model, model_config=model_config, sigma=sigma)



from typing import Tuple

import torch


def sample_confidence_interval(samples: torch.Tensor, alphas: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Computes the (1 - alpha) confidence interval for an array of samples using quantiles.

    Args:
        samples (torch.Tensor): Tensor of shape (..., num_samples).
        alphas (torch.Tensor): Confidence level (e.g., 0.05 for a 95% confidence interval) with shape (num_alphas,).

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: (lower_bound, upper_bound), each of shape (num_alphas, ...)
    """
    threshold = 0.01

    # Compute quantiles
    lower_q = alphas / 2
    upper_q = 1 - alphas / 2

    orig_shape = samples.shape
    samples = samples.reshape(-1, orig_shape[-1])  # Flatten the last dimension

    lower_bound = torch.quantile(samples, lower_q.to(samples.dtype), dim=-1, interpolation="linear")
    upper_bound = torch.quantile(samples, upper_q.to(samples.dtype), dim=-1, interpolation="linear")

    # Handle array of small alphas
    zero_mask = alphas <= threshold
    if zero_mask.any():
        # First reshape to ensure proper indexing
        reshaped_lower = lower_bound.view(len(alphas), -1)
        reshaped_upper = upper_bound.view(len(alphas), -1)

        # Replace zeros with infinities
        reshaped_lower[zero_mask] = float("-inf")
        reshaped_upper[zero_mask] = float("inf")

    lower_bound = lower_bound.reshape(len(alphas), *orig_shape[:-1])
    upper_bound = upper_bound.reshape(len(alphas), *orig_shape[:-1])

    return lower_bound, upper_bound



import math

import torch
import torch.nn.functional as F
from torch import nn


class RMSNorm(torch.nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x):
        output = self._norm(x.float()).type_as(x)
        return output * self.weight


class TransformerEncoderLayer(nn.Module):
    def __init__(self, embed_dim, num_heads, ff_dim):
        super().__init__()
        self.embed_dim = embed_dim
        self.head_dim = embed_dim // num_heads
        self.num_heads = num_heads
        self.kv_proj = nn.Linear(embed_dim, 2 * embed_dim, bias=False)
        self.q_proj = nn.Linear(embed_dim, 3 * embed_dim, bias=False)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=False)

        norm = lambda x: RMSNorm(x)
        self.attn_norm = norm(embed_dim)
        self.ff_norm = norm(embed_dim)
        self.q_norm = norm(self.head_dim)
        self.k_norm = norm(self.head_dim)

    def forward(self, x, context_length):
        x = x.transpose(0, 1)
        B, L, _ = x.size()
        h = self.attn_norm(x)
        q, ff_h, ff_gate = self.q_proj(h).chunk(3, dim=-1)
        k, v = self.kv_proj(h[:, :context_length]).chunk(2, dim=-1)

        q = q.view(B, L, self.num_heads, self.head_dim).transpose(1, 2)
        k = k.view(B, context_length, self.num_heads, self.head_dim).transpose(1, 2)
        v = v.view(B, context_length, self.num_heads, self.head_dim).transpose(1, 2)
        q, k = self.q_norm(q), self.k_norm(k)
        q = q * math.log2(context_length) / 10
        q = self.q_norm(q).to(v.dtype)
        k = self.k_norm(k).to(v.dtype)

        attn = F.scaled_dot_product_attention(q, k, v).transpose(1, 2)
        attn = attn.reshape(B, L, self.num_heads * self.head_dim)
        ff_x = ff_h * F.silu(ff_gate)
        residual = self.ff_norm(self.out_proj(attn + ff_x))
        x_out = x + residual
        return x_out.transpose(0, 1)



import torch

from .model import TabDPTLongContextModel


class DictToObject:
    def __init__(self, d):
        for key, value in d.items():
            if isinstance(value, dict):
                value = DictToObject(value)
            setattr(self, key, value)


def object_to_dict(obj):
    if isinstance(obj, dict):
        return {k: object_to_dict(v) for k, v in obj.items()}
    elif hasattr(obj, "__dict__"):
        return {k: object_to_dict(v) for k, v in vars(obj).items()}
    elif isinstance(obj, list):
        return [object_to_dict(item) for item in obj]
    else:
        return obj


def load_pretrained_tabdpt_model(ckpt_path: str | None, ckpt: dict | None = None) -> TabDPTLongContextModel:
    assert ckpt_path is not None or ckpt is not None, "Either ckpt_path or ckpt must be provided."
    checkpoint = torch.load(ckpt_path, weights_only=False, map_location="cpu") if ckpt is None else ckpt
    config = DictToObject(checkpoint["cfg"])
    model = TabDPTLongContextModel.load(
        model_state=checkpoint["model"],
        config=config,
    )
    return model


def load_pretrained_tabdpt_config(ckpt_path: str | None, ckpt: dict | None = None) -> dict:
    assert ckpt_path is not None or ckpt is not None, "Either ckpt_path or ckpt must be provided."
    checkpoint = torch.load(ckpt_path, weights_only=False, map_location="cpu") if ckpt is None else ckpt
    return object_to_dict(checkpoint["cfg"])
